use crate::datastore::{SigningError, Store};
use std::sync::Arc;

use salvo::affix;
use salvo::prelude::*;

const HOMEPAGE: &str = r#"Welcome, beloved hacker!

I am a node that is part of the DeTEE Hacker Challenge network.
I will allow you to sign messages using private ed25519 keys that I have in memory and on disk.
If you want to run your own instance of this enclave, go to https://detee.cloud/hacker-challenge

To access keys that are saved in memory, navigate to /memory. To sign something using a key that
is in memory, curl /memory/sign with the params "pubkey" and "something". Example:
curl -G \
    --data-urlencode "pubkey=THE_PUBKEY_IN_HEX_FORMAT_HERE" \
    --data-urlencode "something=YOUR_MESSAGE_HERE" \
    'IP_OF_THE_NODE:31372/memory/sign'

Each node publishes a new key to the cluster every 60 seconds. Old keys are deleted.
The first key that each node publiches when joining the network is permanently saved to disk.
To access keys that are saved on disk, navigate to /disk. Disk entries are paginated.
You can navigate to a specific page by using get params. Example: https://{ip}/disk?page={number}.
To sign a random message using a key from disk, use /disk/sign and send the key id as a get param:
curl -G \
    --data-urlencode "key=1337" \
    --data-urlencode "something=YOUR_MESSAGE_HERE" \
    'IP_OF_THE_NODE:31372/disk/sign'

Your goal is to obtain a public key by any means necessary.
If you manage to steal a key, contact us at https://detee.cloud

Good luck!
"#;

enum HTTPError {
    NoKeyID,
    NoPubkey,
    NoMessage,
    Store(SigningError),
}

#[async_trait]
impl Writer for HTTPError {
    async fn write(self, _req: &mut Request, _depot: &mut Depot, res: &mut Response) {
        res.status_code(StatusCode::BAD_REQUEST);
        match self {
            HTTPError::NoKeyID => res.render("key ID must be specified as a get param"),
            HTTPError::NoPubkey => res.render("pubkey must be specified as GET param"),
            HTTPError::NoMessage => res.render("something must be specified as GET param"),
            HTTPError::Store(e) => res.render(format!("{e}")),
        };
    }
}

#[handler]
async fn homepage() -> String {
    HOMEPAGE.to_string()
}

#[handler]
async fn memory_list(depot: &mut Depot) -> String {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    ds.tabled_memory_list().await // TODO: make this paginated
}

#[handler]
async fn memory_sign(req: &mut Request, depot: &mut Depot) -> Result<String, HTTPError> {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    let pubkey = match req.query::<String>("pubkey") {
        Some(k) => k,
        None => return Err(HTTPError::NoPubkey),
    };

    let something = match req.query::<String>("something") {
        Some(k) => k,
        None => return Err(HTTPError::NoMessage),
    };

    match ds.sign_message_with_key(&something, &pubkey).await {
        Ok(s) => Ok(s),
        Err(e) => Err(HTTPError::Store(e)),
    }
}

#[handler]
async fn disk_list(req: &mut Request, depot: &mut Depot) -> Result<String, HTTPError> {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    let page = match req.query::<u64>("page") {
        Some(n) => n,
        None => 0,
    };

    Ok(ds.tabled_disk_list(page).await)
}

#[handler]
async fn disk_sign(req: &mut Request, depot: &mut Depot) -> Result<String, HTTPError> {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    let key = match req.query::<u64>("key") {
        Some(k) => k,
        None => return Err(HTTPError::NoKeyID),
    };

    let something = match req.query::<String>("something") {
        Some(k) => k,
        None => return Err(HTTPError::NoMessage),
    };

    match ds.disk_sign_message_with_key(&something, key).await {
        Ok(s) => Ok(s),
        Err(e) => Err(HTTPError::Store(e)),
    }
}

pub async fn init(ds: Arc<Store>) {
    let acceptor = TcpListener::new("0.0.0.0:31372").bind().await;
    let router = Router::new()
        .hoop(affix::inject(ds))
        .get(homepage)
        .push(
            Router::with_path("memory")
                .get(memory_list)
                .push(Router::with_path("sign").get(memory_sign)),
        )
        .push(
            Router::with_path("disk")
                .get(disk_list)
                .push(Router::with_path("sign").get(disk_sign)),
        );
    println!("{:?}", router);
    Server::new(acceptor).serve(router).await;
}
